from .base import LottieObject, LottieProp, PseudoBool
from .properties import Value, MultiDimensional, ColorValue
from ..nvector import NVector
from ..utils.color import Color


#5: EffectsManager,
#11: MaskEffect,
class EffectValue(LottieObject):
    """!
    Value for an effect
    """
    ## %Effect value type.
    type = None
    _classses = {}

    _props = [
        LottieProp("effect_index", "ix", int, False),
        #LottieProp("match_name", "mn", str, False),
        LottieProp("name", "nm", str, False),
        LottieProp("type", "ty", int, False),
    ]

    def __init__(self):
        ## Effect Index. Used for expressions.
        self.effect_index = None
        ## After Effect's Name. Used for expressions.
        self.name = None

        """
        ## After Effect's Match Name. Used for expressions.
        self.match_name = ""
        """

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not EffectValue._classses:
            EffectValue._classses = {
                sc.type: sc
                for sc in EffectValue.__subclasses__()
            }
        return EffectValue._classses[lottiedict["ty"]]

    def __str__(self):
        return self.name or super().__str__()


## @ingroup Lottie
class Effect(LottieObject):
    """!
    Layer effect
    """
    ## %Effect type.
    type = None
    _classses = {}

    _props = [
        LottieProp("effect_index", "ix", int, False),
        #LottieProp("match_name", "mn", str, False),
        LottieProp("name", "nm", str, False),
        LottieProp("type", "ty", int, False),
        LottieProp("effects", "ef", EffectValue, True),
    ]
    _effects = []

    def __init__(self, *args, **kwargs):
        ## Effect Index. Used for expressions.
        self.effect_index = None
        ## After Effect's Name. Used for expressions.
        self.name = None
        ## Effect parameters
        self.effects = self._load_values(*args, **kwargs)

        """
        ## After Effect's Match Name. Used for expressions.
        self.match_name = ""
        """

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not Effect._classses:
            Effect._classses = {
                sc.type: sc
                for sc in Effect.__subclasses__()
            }
        type = lottiedict["ty"]

        if type in Effect._classses:
            return Effect._classses[type]
        else:
            return Effect

    def _load_values(self, *args, **kwargs):
        values = []
        for i, (name, type) in enumerate(self._effects):
            val = []
            if len(args) > i:
                val = [args[i]]
            if name in kwargs:
                val = [kwargs[name]]
            values.append(type(*val))
        return values

    def __getattr__(self, key):
        for i, (name, type) in enumerate(self._effects):
            if name == key:
                return self.effects[i].value
        return super().__getattr__(key)

    def __str__(self):
        return self.name or super().__str__()


## @ingroup Lottie
## @ingroup LottieCheck
class EffectNoValue(EffectValue):
    _props = []


## @ingroup Lottie
class EffectValueSlider(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 0

    def __init__(self, value=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value(value)


## @ingroup Lottie
class EffectValueAngle(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 1

    def __init__(self, angle=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value(angle)


## @ingroup Lottie
class EffectValueColor(EffectValue):
    _props = [
        LottieProp("value", "v", ColorValue, False),
    ]
    ## %Effect type.
    type = 2

    def __init__(self, value=Color(0, 0, 0)):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = ColorValue(value)


## @ingroup Lottie
class EffectValuePoint(EffectValue):
    _props = [
        LottieProp("value", "v", MultiDimensional, False),
    ]
    ## %Effect type.
    type = 3

    def __init__(self, value=NVector(0, 0)):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = MultiDimensional(value)


## @ingroup Lottie
class EffectValueCheckbox(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 4

    def __init__(self, value=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value(value)


## @ingroup Lottie
## @ingroup LottieCheck
## Lottie-web ignores these
class IgnoredValue(EffectValue):
    _props = [
        LottieProp("value", "v", float, False),
    ]
    ## %Effect type.
    type = 6

    def __init__(self, value=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = value


## @ingroup Lottie
## @ingroup LottieCheck
class EffectValueDropDown(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 7

    def __init__(self, value=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value(value)


## @ingroup Lottie
## @ingroup LottieCheck
class EffectValueLayer(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 10

    def __init__(self):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value()


## @ingroup Lottie
class FillEffect(Effect):
    """!
    Replaces the whole layer with the given color
    @note Opacity is in [0, 1]
    """
    _effects = [
        ("00", EffectValuePoint),
        ("01", EffectValueDropDown),
        ("color", EffectValueColor),
        ("03", EffectValueDropDown),
        ("04", EffectValueSlider),
        ("05", EffectValueSlider),
        ("opacity", EffectValueSlider),
    ]
    ## %Effect type.
    type = 21


## @ingroup Lottie
class StrokeEffect(Effect):
    _effects = [
        ("00", EffectValueColor),
        ("01", EffectValueCheckbox),
        ("02", EffectValueCheckbox),
        ("color", EffectValueColor),
        ("04", EffectValueSlider),
        ("05", EffectValueSlider),
        ("06", EffectValueSlider),
        ("07", EffectValueSlider),
        ("08", EffectValueSlider),
        ("09", EffectValueDropDown),
        ("type", EffectValueDropDown),
    ]
    ## %Effect type.
    type = 22


## @ingroup Lottie
class TritoneEffect(Effect):
    """!
    Maps layers colors based on bright/mid/dark colors
    """
    _effects = [
        ("bright", EffectValueColor),
        ("mid", EffectValueColor),
        ("dark", EffectValueColor),
    ]
    ## %Effect type.
    type = 23


"""
## @ingroup Lottie
## @ingroup LottieCheck
class GroupEffect(Effect):
    _props = [
        LottieProp("enabled", "en", PseudoBool, False),
    ]

    def __init__(self):
        Effect.__init__(self)
        ## Enabled AE property value
        self.enabled = True
"""


## @ingroup Lottie
## @ingroup LottieCheck
class ProLevelsEffect(Effect):
    _effects = [
        ("00", EffectValueDropDown),
        ("01", EffectNoValue),
        ("02", EffectNoValue),
        ("comp_inblack", EffectValueSlider),
        ("comp_inwhite", EffectValueSlider),
        ("comp_gamma", EffectValueSlider),
        ("comp_outblack", EffectValueSlider),
        ("comp_outwhite", EffectNoValue),
        ("08", EffectNoValue),
        ("09", EffectValueSlider),
        ("r_inblack", EffectValueSlider),
        ("r_inwhite", EffectValueSlider),
        ("r_gamma", EffectValueSlider),
        ("r_outblack", EffectValueSlider),
        ("r_outwhite", EffectNoValue),
        ("15", EffectValueSlider),
        ("16", EffectValueSlider),
        ("g_inblack", EffectValueSlider),
        ("g_inwhite", EffectValueSlider),
        ("g_gamma", EffectValueSlider),
        ("g_outblack", EffectValueSlider),
        ("g_outwhite", EffectNoValue),
        ("22", EffectValueSlider),
        ("b3", EffectValueSlider),
        ("b_inblack", EffectValueSlider),
        ("b_inwhite", EffectValueSlider),
        ("b_gamma", EffectValueSlider),
        ("b_outblack", EffectValueSlider),
        ("b_outwhite", EffectNoValue),
        ("29", EffectValueSlider),
        ("a_inblack", EffectValueSlider),
        ("a_inwhite", EffectValueSlider),
        ("a_gamma", EffectValueSlider),
        ("a_outblack", EffectValueSlider),
        ("a_outwhite", EffectNoValue),
    ]
    ## %Effect type.
    type = 24


## @ingroup Lottie
class TintEffect(Effect):
    """!
    Colorizes the layer
    @note Opacity is in [0, 100]
    """
    _effects = [
        ("color_black", EffectValueColor),
        ("color_white", EffectValueColor),
        ("opacity", EffectValueSlider),
    ]
    ## %Effect type.
    type = 20


## @ingroup Lottie
class DropShadowEffect(Effect):
    """!
    Adds a shadow to the layer
    @note Opacity is in [0, 255]
    """
    _effects = [
        ("color", EffectValueColor),
        ("opacity", EffectValueSlider),
        ("angle", EffectValueAngle),
        ("distance", EffectValueSlider),
        ("blur", EffectValueSlider),
    ]
    ## %Effect type.
    type = 25


## @ingroup Lottie
## @ingroup LottieCheck
class Matte3Effect(Effect):
    _effects = [
        ("index", EffectValueSlider),
    ]
    ## %Effect type.
    type = 28


## @ingroup Lottie
class GaussianBlurEffect(Effect):
    """!
    Gaussian blur
    """
    _effects = [
        ("sigma", EffectValueSlider),
        ("dimensions", EffectValueSlider),
        ("wrap", EffectValueCheckbox),
    ]
    ## %Effect type.
    type = 29


#class ChangeColorEffect(Effect):
    #"""!
    #Gaussian blur
    #"""
    #_effects = [
        #("view", EffectValueDropDown),
        #("hue", EffectValueSlider),
        #("lightness", EffectValueSlider),
        #("saturation", EffectValueSlider),
        #("color_to_change", EffectValueColor),
        #("tolerance", EffectValueSlider),
        #("softness", EffectValueSlider),
        #("match", EffectValueDropDown),
        #("invert_mask", EffectValueDropDown),
    #]
    ### %Effect type.
    #type = 29


## @ingroup Lottie
class ChangeToColorEffect(Effect):
    """!
    Change to color
    """
    _effects = [
        ("from_color", EffectValueColor),
        ("to_color", EffectValueColor),
        ("change", EffectValueDropDown),
        ("change_by", EffectValueDropDown),
        ("tolerance", IgnoredValue),
        ("hue", EffectValueSlider),
        ("lightness", EffectValueSlider),
        ("saturation", EffectValueSlider),
        ("saturation_", IgnoredValue),
        ("softness", EffectValueSlider),
        ("view_correction", EffectValueDropDown),
    ]
    ## %Effect type.
    type = 5



import warnings
from .base import LottieObject, LottieProp, PseudoBool, LottieEnum
from .effects import Effect
from .helpers import Transform, Mask
from .shapes import ShapeElement
from .text import TextAnimatorData
from .properties import Value


## @ingroup Lottie
class BlendMode(LottieEnum):
    Normal = 0
    Multiply = 1
    Screen = 2
    Overlay = 3
    Darken = 4
    Lighten = 5
    ColorDodge = 6
    ColorBurn = 7
    HardLight = 8
    SoftLight = 9
    Difference = 10
    Exclusion = 11
    Hue = 12
    Saturation = 13
    Color = 14
    Luminosity = 15


## @ingroup Lottie
## @todo SVG masks
class MatteMode(LottieEnum):
    Normal = 0
    Alpha = 1
    InvertedAlpha = 2
    Luma = 3
    InvertedLuma = 4


## @ingroup Lottie
class Layer(LottieObject):
    _props = [
        LottieProp("threedimensional", "ddd", PseudoBool, False),
        LottieProp("hidden", "hd", bool, False),
        LottieProp("type", "ty", int, False),
        LottieProp("name", "nm", str, False),
        LottieProp("parent_index", "parent", int, False),

        LottieProp("stretch", "sr", float, False),
        LottieProp("transform", "ks", Transform, False),
        LottieProp("auto_orient", "ao", PseudoBool, False),

        LottieProp("in_point", "ip", float, False),
        LottieProp("out_point", "op", float, False),
        LottieProp("start_time", "st", float, False),
        LottieProp("blend_mode", "bm", BlendMode, False),

        LottieProp("matte_mode", "tt", MatteMode, False),
        LottieProp("index", "ind", int, False),
        #LottieProp("css_class", "cl", str, False),
        #LottieProp("layer_html_id", "ln", str, False),
        LottieProp("has_masks", "hasMask", bool, False),
        LottieProp("masks", "masksProperties", Mask, True),
        LottieProp("effects", "ef", Effect, True),
    ]
    ## %Layer type.
    ## @see https://github.com/bodymovin/bodymovin-extension/blob/master/bundle/jsx/enums/layerTypes.jsx
    type = None
    _classses = {}

    @property
    def has_masks(self):
        """!
        Whether the layer has some masks applied
        """
        return bool(self.masks) if getattr(self, "masks") is not None else None

    def __init__(self):
        ## Transform properties
        self.transform = Transform()
        ## Auto-Orient along path AE property.
        self.auto_orient = False
        ## 3d layer flag
        self.threedimensional = False
        ## Hidden layer
        self.hidden = None
        ## Layer index in AE. Used for parenting and expressions.
        self.index = None

        """
        # Parsed layer name used as html class on SVG/HTML renderer
        #self.css_class = ""
        # Parsed layer name used as html id on SVG/HTML renderer
        #self.layer_html_id = ""
        """
        ## In Point of layer. Sets the initial frame of the layer.
        self.in_point = None
        ## Out Point of layer. Sets the final frame of the layer.
        self.out_point = None
        ## Start Time of layer. Sets the start time of the layer.
        self.start_time = 0
        ## After Effects Layer Name. Used for expressions.
        self.name = None
        ## List of Effects
        self.effects = None
        ## Layer Time Stretching
        self.stretch = 1
        ## Layer Parent. Uses ind of parent.
        self.parent_index = None
        ## List of Masks
        self.masks = None
        ## Blend Mode
        self.blend_mode = BlendMode.Normal
        ## Matte mode, the layer will inherit the transparency from the layer above
        self.matte_mode = None
        ## Composition owning the layer, set by add_layer
        self.composition = None

    def add_child(self, layer):
        if not self.composition or self.index is None:
            raise Exception("Must set composition / index first")
        self._child_inout_auto(layer)
        self.composition.add_layer(layer)
        layer.parent_index = self.index
        return layer

    def _child_inout_auto(self, layer):
        if layer.in_point is None:
            layer.in_point = self.in_point
        if layer.out_point is None:
            layer.out_point = self.out_point

    @property
    def parent(self):
        if self.parent_index is None:
            return None
        return self.composition.layer(self.parent_index)

    @parent.setter
    def parent(self, layer):
        if layer is None:
            self.parent_index = None
        else:
            self.parent_index = layer.index
            layer._child_inout_auto(self)

    @property
    def children(self):
        for layer in self.composition.layers:
            if layer.parent_index == self.index:
                yield layer

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not Layer._classses:
            Layer._classses = {
                sc.type: sc
                for sc in Layer.__subclasses__()
            }
        type_id = lottiedict["ty"]
        if type_id not in Layer._classses:
            warnings.warn("Unknown layer type: %s" % type_id)
            return Layer
        return Layer._classses[type_id]

    def __repr__(self):
        return "<%s %s %s>" % (type(self).__name__, self.index, self.name)

    def __str__(self):
        return "%s %s" % (
            self.name or super().__str__(),
            self.index if self.index is not None else ""
        )

    def remove(self):
        """!
        @brief Removes this layer from the componsitin
        """
        self.composition.remove_layer(self)


## @ingroup Lottie
class NullLayer(Layer):
    """!
    Layer with no data, useful to group layers together
    """
    ## %Layer type.
    type = 3

    def __init__(self):
        Layer.__init__(self)


## @ingroup Lottie
class TextLayer(Layer):
    _props = [
        LottieProp("data", "t", TextAnimatorData, False),
    ]
    ## %Layer type.
    type = 5

    def __init__(self):
        Layer.__init__(self)
        ## Text Data
        self.data = TextAnimatorData()


## @ingroup Lottie
class ShapeLayer(Layer):
    """!
    Layer containing ShapeElement objects
    """
    _props = [
        LottieProp("shapes", "shapes", ShapeElement, True),
    ]
    ## %Layer type.
    type = 4

    def __init__(self):
        Layer.__init__(self)
        ## Shape list of items
        self.shapes = [] # ShapeElement

    def add_shape(self, shape):
        self.shapes.append(shape)
        return shape

    def insert_shape(self, index, shape):
        self.shapes.insert(index, shape)
        return shape


## @ingroup Lottie
## @todo SIF I/O
class ImageLayer(Layer):
    _props = [
        LottieProp("image_id", "refId", str, False),
    ]
    ## %Layer type.
    type = 2

    def __init__(self, image_id=""):
        Layer.__init__(self)
        ## id pointing to the source image defined on 'assets' object
        self.image_id = image_id


## @ingroup Lottie
class PreCompLayer(Layer):
    _props = [
        LottieProp("reference_id", "refId", str, False),
        LottieProp("time_remapping", "tm", Value, False),
        LottieProp("width", "w", int, False),
        LottieProp("height", "h", int, False),
    ]
    ## %Layer type.
    type = 0

    def __init__(self, reference_id=""):
        Layer.__init__(self)
        ## id pointing to the source composition defined on 'assets' object
        self.reference_id = reference_id
        ## Comp's Time remapping
        self.time_remapping = None
        ## Width
        self.width = 512
        ## Height
        self.height = 512


## @ingroup Lottie
class SolidColorLayer(Layer):
    """!
    Layer with a solid color rectangle
    """
    _props = [
        LottieProp("color", "sc", str, False),
        LottieProp("height", "sh", float, False),
        LottieProp("width", "sw", float, False),
    ]
    ## %Layer type.
    type = 1

    def __init__(self, color="", width=512, height=512):
        Layer.__init__(self)
        ## Color of the layer as a @c \#rrggbb hex
        # @todo Convert NVector to string
        self.color = color
        ## Height of the layer.
        self.height = height
        ## Width of the layer.
        self.width = width



# SPDX-FileCopyrightText: 2019-2023 Mattia Basaglia <dev@dragon.best>
# SPDX-License-Identifier: GPL-3.0-or-later
def main(window, document, settings):
    dialog = window.create_dialog("create_frames.ui")

    if not dialog:
        return

    dialog.set_value("spin_first", "minimum", document.main.animation.first_frame)
    dialog.set_value("spin_first", "maximum", document.main.animation.last_frame)

    dialog.set_value("spin_last", "minimum", document.main.animation.first_frame)
    dialog.set_value("spin_last", "maximum", document.main.animation.last_frame)

    dialog.set_value("spin_first", "value", document.main.animation.first_frame)
    dialog.set_value("spin_last", "value", document.main.animation.last_frame)

    initial = (document.main.animation.last_frame - document.main.animation.first_frame) / document.main.fps * 24
    dialog.set_value("spin_count", "value", int(initial))

    if dialog.exec():
        first = dialog.get_value("spin_first", "value")
        last = dialog.get_value("spin_last", "value")
        count = dialog.get_value("spin_count", "value")

        if first == last:
            return

        if first > last:
            (last, first) = (first, last)

        duration = (last - first) / count
        with document.macro("Create frame layers"):
            for i in range(count):
                layer = window.current_composition.add_shape("Layer")
                layer.name = "Frame %s" % i
                layer.animation.first_frame = duration * i
                layer.animation.last_frame = duration * (i+1)



# NOTE: requires pillow, pypotrace>=0.2, numpy, scipy to be installed
from PIL import Image
import potrace
import numpy
import enum
from scipy.cluster.vq import kmeans
from .. import objects
from ..nvector import NVector
from .pixel import _vectorizing_func


class QuanzationMode(enum.Enum):
    Nearest = 1
    Exact = 2


class RasterImage:
    def __init__(self, data):
        self.data = data

    @classmethod
    def from_pil(cls, image):
        return cls(numpy.array(image))

    #@classmethod
    #def open(cls, filename):
        #return cls.from_pil(Image.open(filename))

    def k_means(self, n_colors):
        """!
        Returns a list of centroids
        """
        colors = []
        for row in range(self.data.shape[0]):
            for column in range(self.data.shape[1]):
                if self.get_alpha(row, column) == 255:
                    colors.append(self.data[row][column])

        colors = numpy.array(colors, numpy.float)
        return kmeans(colors, n_colors+1)[0]

    def get_alpha(self, row, column):
        if self.data.shape[2] >= 4:
            return self.data[row][column][3]
        return 255

    def quantize(self, codebook, quantization_mode=QuanzationMode.Nearest):
        """!
        Returns a list of tuple [color, data] where for each color in codebook
        data is a bit mask for the image

        You can get codebook from k_means
        """
        if codebook is None or len(codebook) == 0:
            return [(numpy.array([0., 0., 0., 255.]), self.mono())]

        mono_data = []
        for c in codebook:
            mono_data.append((c, numpy.zeros(self.data.shape[:2])))

        for row in range(self.data.shape[0]):
            for column in range(self.data.shape[1]):
                if self.get_alpha(row, column) == 255:
                    if quantization_mode == QuanzationMode.Nearest:
                        min_norm = 511 # (norm of [255, 255, 255, 255]) + 1
                        best = None
                        for color, bitmap in mono_data:
                            norm = numpy.linalg.norm(self.data[row][column] - color)
                            if norm < min_norm:
                                min_norm = norm
                                best = bitmap
                                if norm == 0:
                                    break
                        best[row][column] = 1
                    else:
                        for color, bitmap in mono_data:
                            if numpy.array_equal(color, self.data[row][column]):
                                bitmap[row][column] = 1
                                break

        return mono_data

    def mono(self):
        """!
        Returns a bit mask of opaque pixels
        """
        mono_data = numpy.zeros(self.data.shape[:2])
        for row in range(self.data.shape[0]):
            for column in range(self.data.shape[1]):
                mono_data[row][column] = int(self.data[row][column][3] == 255)
        return mono_data


class Vectorizer:
    def __init__(self):
        self.palette = None
        self.layers = {}

    def _create_layer(self, animation, layer_name):
        layer = animation.add_layer(objects.ShapeLayer())
        if layer_name:
            self.layers[layer_name] = layer
            layer.name = layer_name
        return layer

    def prepare_layer(self, animation, layer_name=None):
        layer = self._create_layer(animation, layer_name)
        layer._max_verts = {}
        if self.palette is None:
            group = layer.add_shape(objects.Group())
            group.name = "bitmap"
            layer._max_verts[group.name] = 0
            group.add_shape(objects.Path())
            group.add_shape(objects.Fill(NVector(0, 0, 0)))
        else:
            for color in self.palette:
                group = layer.add_shape(objects.Group())
                group.name = "color_%s" % "".join("%02x" % int(c) for c in color)
                layer._max_verts[group.name] = 0
                fcol = color/255
                fill = group.add_shape(objects.Fill(NVector(*fcol)))
                if len(fcol) > 3 and fcol[3] < 1:
                    fill.opacity.value = fcol[3] * 100
        return layer

    def raster_to_layer(self, animation, raster, layer_name=None, mode=QuanzationMode.Nearest):
        layer = self.prepare_layer(animation, layer_name)
        mono_data = raster.quantize(self.palette, mode)
        for (color, bitmap), group in zip(mono_data, layer.shapes):
            self.raster_to_shapes(group, bitmap)
        return layer

    def raster_to_shapes(self, group, mono_data):
        shapes = []
        for bezier in self.raster_to_bezier(mono_data):
            shape = group.insert_shape(0, objects.Path())
            shapes.append(shape)
            shape.shape.value = bezier
        return shapes

    def raster_to_bezier(self, mono_data):
        bmp = potrace.Bitmap(mono_data)
        path = bmp.trace()
        shapes = []
        for curve in path:
            bezier = objects.Bezier()
            shapes.append(bezier)
            bezier.add_point(NVector(*curve.start_point))
            for segment in curve:
                if segment.is_corner:
                    bezier.add_point(NVector(*segment.c))
                    bezier.add_point(NVector(*segment.end_point))
                else:
                    sp = NVector(*bezier.vertices[-1])
                    ep = NVector(*segment.end_point)
                    c1 = NVector(*segment.c1) - sp
                    c2 = NVector(*segment.c2) - ep
                    bezier.out_tangents[-1] = c1
                    bezier.add_point(ep, c2)
        return shapes

    def _frame_keyframe(self, layer, group, time, shapes, beziers):
        if shapes:
            # TODO handle multiple shapes
            nverts = len(beziers[0].vertices)
            if nverts > layer._max_verts[group.name]:
                layer._max_verts[group.name] = nverts
            for shape, bezier in zip(shapes, beziers):
                shape.shape.add_keyframe(time, bezier)

    def raster_to_frame(self, animation, raster, layer_name, time, mode=QuanzationMode.Nearest):
        mono_data = raster.quantize(self.palette, mode)

        if layer_name not in self.layers:
            layer = self.prepare_layer(animation, layer_name)
            for (color, bitmap), group in zip(mono_data, layer.shapes):
                shapes = self.raster_to_shapes(group, bitmap)
                beziers = [s.shape.value for s in shapes]
                self._frame_keyframe(layer, group, time, shapes, beziers)
        else:
            layer = self.layers[layer_name]

            for (color, bitmap), group in zip(mono_data, layer.shapes):
                shapes = [s for s in group.shapes if isinstance(s, objects.Path)]
                beziers = self.raster_to_bezier(bitmap)
                self._frame_keyframe(layer, group, time, shapes, beziers)

    def adjust_missing_vertices(self, layer_name):
        layer = self.layers[layer_name]
        for group in layer.shapes:
            # TODO handle multiple shapes
            shape = group.shapes[0]
            nverts = layer._max_verts[group.name]
            if shape.shape.animated:
                for kf in shape.shape.keyframes:
                    bezier = kf.start
                    count = nverts - len(bezier.vertices)
                    bezier.vertices += [bezier.vertices[-1]] * count
                    bezier.in_tangents += [NVector(0, 0)] * count
                    bezier.out_tangents += [NVector(0, 0)] * count

    def duplicate_start_frame(self, layer_name, time):
        layer = self.layers[layer_name]
        for group in layer.shapes:
            shape = group.shapes[0]
            bezier = shape.shape.keyframes[0].start
            group.shapes[0].shape.add_keyframe(time, bezier)


def color2numpy(vcolor):
    l = (vcolor * 255).components
    if len(l) == 3:
        l.append(255)
    return numpy.array(l, numpy.uint8)


def raster_to_animation(filenames, n_colors=1, frame_delay=1,
                        looping=True, framerate=60, palette=[],
                        mode=QuanzationMode.Nearest):

    vc = Vectorizer()

    def callback(animation, raster, frame):
        raster = RasterImage.from_pil(raster)
        if vc.palette is None:
            if palette:
                vc.palette = [color2numpy(c) for c in palette]
            elif n_colors > 1:
                vc.palette = raster.k_means(n_colors)
        #vc.raster_to_frame(animation, raster, "anim", frame * frame_delay, mode)
        layer = vc.raster_to_layer(animation, raster, "frame_%s" % frame, mode)
        layer.in_point = frame * frame_delay
        layer.out_point = (frame + 1) * frame_delay

    animation = _vectorizing_func(filenames, frame_delay, framerate, callback)

    #vc.adjust_missing_vertices("anim")
    #if looping and animation._nframes > 1:
        #animation.out_point += frame_delay
        #vc.duplicate_start_frame("anim", animation.out_point)
    #elif animation._nframes == 1:
        #for g in animation.find("anim").shapes:
            #for shape in g.find_all(objects.Path):
                #shape.shape.clear_animation(shape.shape.get_value(0))

    #animation.find("anim").out_point = animation.out_point

    return animation



# SPDX-FileCopyrightText: 2019-2023 Mattia Basaglia <dev@dragon.best>
# SPDX-License-Identifier: GPL-3.0-or-later
import glaxnimate


class HueShiftVisitor(glaxnimate.model.Visitor):
    def __init__(self, amount, times, easing):
        super().__init__()
        self.amount = amount
        self.gradients = set()
        self.named_colors = set()
        self.times = times
        self.easing = easing

    def _color(self, color, amount):
        return glaxnimate.utils.Color.from_hsv(
            (color.hue + amount) % 360,
            color.saturation,
            color.value,
            color.alpha
        )

    def _prop_color(self, value, amount):
        if isinstance(value, list):
            return [
                (offset, self._color(color, amount))
                for offset, color in value
            ]
        return self._color(value, amount)

    def set_time(self, time):
        self.gradients = set()
        self.named_colors = set()
        self.time = time

    def _set_value(self, prop):
        value = prop.value
        if not self.times:
            prop.value = self._prop_color(value, self.amount)
        else:
            amount = 0
            for time in self.times:
                kf = prop.set_keyframe(time, self._prop_color(value, amount))
                kft = kf.transition
                kft.before_descriptive = self.easing
                kft.after_descriptive = self.easing
                kf.transition = kft
                amount += self.amount

    def on_visit_node(self, node):
        if isinstance(node, glaxnimate.model.shapes.Styler):
            self._set_value(node.color)
            if node.use:
                if isinstance(node.use, glaxnimate.model.assets.NamedColor):
                    if node.use not in self.named_colors:
                        self._set_value(node.use.color)
                        self.named_colors.add(node.use)
                elif isinstance(node.use, glaxnimate.model.assets.Gradient) and node.use.colors:
                    if node.use.colors not in self.gradients:
                        self._set_value(node.use.colors.colors)
                        self.gradients.add(node.use.colors)


def hue_shift(window, document, settings):
    with document.macro("Hue Shift"):
        visitor = HueShiftVisitor(settings["amount"], None, None)
        for shape in window.cleaned_selection:
            visitor.visit(shape, True)


def hue_shift_cycle(window, document, settings):
    with document.macro("Hue Shift Cycle"):
        easing_str = settings["easing"]
        easing = glaxnimate.model.KeyframeTransition.Descriptive.Linear
        if easing_str == "Ease":
            easing = glaxnimate.model.KeyframeTransition.Descriptive.Ease
        elif easing_str == "Hold":
            easing = glaxnimate.model.KeyframeTransition.Descriptive.Hold

        amount = settings["amount"]
        duration = settings["duration"]
        times = [
            document.current_time + step * duration
            for step in range(settings["step_count"])
        ]

        visitor = HueShiftVisitor(amount, times, easing)

        for shape in window.cleaned_selection:
            visitor.visit(shape, True)





